"""
    This module contains the processors used by the pypox framework.

    The processors handle different aspects of the HTTP request and response, such as cookies, headers, JSON data,
    path parameters, query parameters, and websockets.

    Optional dependencies include PydanticProcessor, HTMXProcessor, and JinjaProcessor.
"""
